#include	"stdThread.h"
#include	"SendRecvThread.h"	// 送受信スレッドを使うため
#include	"MySyncObject.h"	// 同期オブジェクトを使うため
#include	"define.h"


// メニューで使用する文字
#define		CMD_QUIT_CHAR				'q'
#define		CMD_SEND_MSG_CHAR			'm'

// 関数の宣言
BOOL CreateSocket(LPCSTR szAddress, WORD wServerPort, WORD wClientPort);	// ★UDPソケットの作成
BOOL DestroySocket(SOCKET &fd);					// 切断とソケットの破棄
void Stop();									// すべてのソケットを破棄する
void DispMenu();								// メニューの表示
int	 GetKeyString(LPSTR pszString, int iSize);	// キーボード入力文字の取得
BOOL KillZombei();								// ゾンビ状態のスレッドを破棄する
BOOL ShowInterfaceIndex(addrinfo *pAddr);		// インターフェースを表示する
BOOL SetHopLimit(SOCKET fd, int iFamily, int iLimit);	// ホップリミットを設定する
BOOL SendMessagePacket(char *pcData, int iSize);	// メッセージパケットの送信

// 変数の宣言
CSendRecvThread		*m_pCSendRecvThread = NULL;
ConnectionInfoRec	*m_pConInfo = NULL;
CMySyncObject		*m_pCMySyncObject = NULL;



int main(int argc, char *argv[])
{
	char	szKeyInBuff[81];
	WORD	wClientPort = 0;		// ★送信元ポート

	m_pCMySyncObject = new CMySyncObject();
	m_pCMySyncObject->Initialize();

	// ★起動パラメータチェック 送信元ポートを指定可能に
	if ((argc != 3) && (argc != 4))
	{
		fprintf(stderr, "Usage:  %s <ServerAddress> <ServerPort> [ClientPort]\n", argv[0]);
		goto L_END;
	}
	if (argc == 4)
		wClientPort = (WORD)atol(argv[3]);

	// ★ソケットの作成 送受信スレッドを開始する
	if (CreateSocket(argv[1], (WORD)atol(argv[2]), wClientPort) == FALSE)
		goto L_END;
	// メニューの表示
	DispMenu();
	while (1)
	{
		if (KillZombei() == TRUE)			// ★致命的エラーならプログラム終了
			break;
		// キーボードから入力された文字列が'q'なら終了
		// 'm'ならメッセージを入力後は送信する		
		switch (GetKeyString(szKeyInBuff, sizeof(szKeyInBuff)- 1))
		{
			case 0:							// 入力なし、何もしない
				break;
			case CMD_QUIT_CHAR:				// 'q'入力終了
				goto L_END;
			case CMD_SEND_MSG_CHAR:			// 'm'入力szKeyInBuffに文字列が格納されている
				// 送信は送信データをセットするだけ、実際の送信はCSendRecvThreadで実施
				// メッセージパケットの送信
				SendMessagePacket(szKeyInBuff, strlen(szKeyInBuff));
				// メニューの表示
				DispMenu();
				break;
		}
	}

L_END:
	// すべてのソケットの破棄
	Stop();
	m_pCMySyncObject->Uninitialize();
	SAFE_DELETE(m_pCMySyncObject)
	return(0);
}


//----------------------------------------------
// function
//		★UDPソケットの作成
//		送受信スレッドを開始する
// parameter
//		LPCSTR	szAddress	[in]送信先アドレス
//		WORD	wServerPort	[in]送信先ポート
//		WORD	wClientPort	[in]送信元ポート/0
// return
//		TRUE/FALSE
//----------------------------------------------
BOOL CreateSocket(LPCSTR szAddress, WORD wServerPort, WORD wClientPort)
{
	BOOL		fRet = TRUE;
	SOCKET		fd = INVALID_SOCKET;
	addrinfo	hints, *pres = NULL, *pTemp = NULL, *pres2 = NULL;
	char		szPort[NI_MAXSERV];
	int 		on, iRet;

	memset(&hints, 0, sizeof(hints));
	hints.ai_socktype = SOCK_DGRAM;			// UDP
	DISABLE_C4996
	sprintf(szPort, "%d", wServerPort);		// ポート番号=サービス
	ENABLE_C4996
	// アドレスを指定してgetaddrinfoを実行
	if (getaddrinfo(szAddress, szPort, &hints, &pres) == 0)
	{
		pTemp = pres;
		while (pTemp != NULL)
		{
			// socketを実行
			if ((fd = socket(pTemp->ai_family, pTemp->ai_socktype, pTemp->ai_protocol)) == INVALID_SOCKET)
				goto L_NEXT;
			// 使用するインターフェースを表示する
			ShowInterfaceIndex(pTemp);
			// HopLimitを取得設定する
			SetHopLimit(fd, pTemp->ai_family, 10);
			// socketが作れたのでこのソケット採用することにして抜ける
			fRet = TRUE;
			break;
L_NEXT:
			pTemp = pTemp->ai_next;
		}
	}
	if (fd == INVALID_SOCKET)		// ソケット作成に失敗
	{
		DispErrorMsg("Err:socket");
		fRet = FALSE;
		goto L_END;
	}
	// 送信元ポートが指定されたときはbindを実施する
	if (wClientPort != 0)
	{
		memset(&hints, 0, sizeof(hints));
		// 使用するアドレス情報と同じ条件でポート番号だけ送信元にする
		hints.ai_family = pTemp->ai_family;
		hints.ai_socktype = pTemp->ai_socktype;
		hints.ai_protocol = pTemp->ai_protocol;
		// すべてのインターフェースで受信可
		hints.ai_flags = AI_PASSIVE;
		DISABLE_C4996
		sprintf(szPort, "%d", wClientPort);			// ポート番号=サービス
		ENABLE_C4996
		if (getaddrinfo(NULL, szPort, &hints, &pres2) == 0)
		{
#ifdef IPV6_V6ONLY
			// IPv6ソケットでIPv4射影アドレスを使用しないように設定
			if (pTemp->ai_family == AF_INET6)
			{
				on = 1;
				if (setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, (char *)&on, sizeof(on)) < 0)
				{
					DispErrorMsg("Err:setsockopt");
					fRet = FALSE;
					goto L_END;
				}
			}
#endif
			// ★UDPではSO_REUSEADDRを使わない
			if ((iRet = bind(fd, pres2->ai_addr, (int)pres2->ai_addrlen)) == SOCKET_ERROR)
			{
				DispErrorMsg("Err:bind");
				fRet = FALSE;
				goto L_END;
			}
		}
	}
	// 送受信スレッドを開始する
	m_pConInfo = (ConnectionInfoRec *)calloc(1, sizeof(ConnectionInfoRec));
	m_pConInfo->fdClient = fd;
	m_pConInfo->ai_addr = (sockaddr *)calloc(pTemp->ai_addrlen, sizeof(BYTE));	// ★
	m_pConInfo->ai_addrlen = pTemp->ai_addrlen;
	memcpy(m_pConInfo->ai_addr, pTemp->ai_addr, pTemp->ai_addrlen);
	m_pConInfo->pCMySyncObject = m_pCMySyncObject;
	m_pCSendRecvThread = new CSendRecvThread(m_pConInfo);
	m_pCSendRecvThread->Begin();

L_END:
	if (pres != NULL)
		freeaddrinfo(pres);
	if (pres2 != NULL)
		freeaddrinfo(pres2);

	if((fRet == FALSE) && (fd != INVALID_SOCKET))
	{
		DestroySocket(fd);
	}
	return(fRet);
}


//----------------------------------------------
// function
//    ソケットの破棄
// parameter
//    SOCKET		&fd[in/out]破棄するソケット
// return
//    TRUE/FALSE
//----------------------------------------------
BOOL DestroySocket(SOCKET &fd)
{
	if (fd != INVALID_SOCKET)
	{
		fprintf(stderr, "DestroySocket()\n");
		shutdown(fd, SHUT_RDWR);     // 受信も送信も停止
		close(fd);
		fd = INVALID_SOCKET;
	}
	return(TRUE);
}

//----------------------------------------------
// function
//    送受信スレッドの終了、ソケットの破棄
// parameter
//    なし
// return
//    なし
//----------------------------------------------
void Stop()
{
	if (m_pCSendRecvThread != NULL)
	{
		fprintf(stderr, "Stop()\n");
		m_pCSendRecvThread->End();
		m_pCSendRecvThread->WaitForEnd();
		SAFE_DELETE(m_pCSendRecvThread)

		DestroySocket(m_pConInfo->fdClient);
		SAFE_FREE(m_pConInfo->ai_addr)			// ★
		SAFE_FREE(m_pConInfo)
	}
}

//----------------------------------------------
// function
//    メニューの表示
// parameter
//    なし
// return
//    なし
//----------------------------------------------
void DispMenu()
{
	Locate(1, 8, 2);
	fprintf(stderr, "%c:quit %c:メッセージ送信 : ",
		CMD_QUIT_CHAR, CMD_SEND_MSG_CHAR);
}

#if 0
//----------------------------------------------
// function
//    キーボード入力(改行まで)の取得
//		入力がないときはすぐにリターン
// parameter
//    LPSTR		pszString	[in/out]データ格納用バッファ
//    int		iSize		[in]バッファのサイズ
// return
//    0:入力なし CMD_QUIT_CHAR:終了 CMD_SEND_MSG_CHAR:メッセージ送信
//----------------------------------------------
int GetKeyString(LPSTR pszString, int iSize)
{
	BOOL			fRet = FALSE;
	int				iRet = 0;

	if (kbhit())
	{
		switch (getchar())
		{
			case CMD_QUIT_CHAR:
				iRet = CMD_QUIT_CHAR;
				break;
			case CMD_SEND_MSG_CHAR:
				fprintf(stderr, "\nInput Msg : ");
				fgets(pszString, iSize, stdin);
				if (strlen(pszString) > 0)
				{
				// 改行まで読み込んでいるので
					if (pszString[strlen(pszString) - 1] == '\n')
						pszString[strlen(pszString) - 1] = 0;
					iRet = CMD_SEND_MSG_CHAR;
				}
				break;
			default:
				break;
		}
	}
	return(iRet);
}
#else
// リモートデバッグではttyじゃないのでkbhitは使えない
// 標準入力は0なのでselectのままでOKですが、練習のためにpollに
int GetKeyString(LPSTR pszString, int iSize)
{
	BOOL			fRet = FALSE;
	int				iRet = 0;
	pollfd			fds[1] = {0};

	fds[0].fd = 0;					// 0:標準入力 1:標準出力 2:標準エラー
	fds[0].events = POLLIN;			// Enterキー入力を対象に
	poll(fds, 1, 10);
	if (fds[0].revents & POLLIN)	// ENTERキー入力があった
	{
		memset(pszString, 0, iSize);
		read(0, pszString, iSize);
		switch (pszString[0])
		{
			case CMD_QUIT_CHAR:
				iRet = CMD_QUIT_CHAR;
				break;
			case CMD_SEND_MSG_CHAR:
				while (1)
				{
					poll(fds, 1, 10);
					if (fds[0].revents & POLLIN)
						read(0, pszString, iSize);
					else
						break;
				}
				fprintf(stderr, "\n送信メッセージを入力してください : ");
				memset(pszString, 0, iSize);
				read(0, pszString, iSize);	// ENTERキーまで読み込む
				if (strlen(pszString) > 0)
				{
				// 改行まで読み込んでいるので
					if (pszString[strlen(pszString) - 1] == '\n')
						pszString[strlen(pszString) - 1] = 0;
					iRet = CMD_SEND_MSG_CHAR;
				}
				break;
			default:
				break;
		}
	}
	return(iRet);
}
#endif


//----------------------------------------------
// function
//	終了済みのCSendRecvThreadを破棄する
// parameter
//	なし
// return
//	TRUE:破棄した/FALSE:破棄すべきものがなかった
//----------------------------------------------
BOOL KillZombei()
{
	BOOL	fRet = FALSE;

	if (m_pCSendRecvThread != NULL)
	{
		if (m_pCSendRecvThread->IsZombie() == TRUE)
		{
			m_pCSendRecvThread->End();
			m_pCSendRecvThread->WaitForEnd();
			SAFE_DELETE(m_pCSendRecvThread);

			DestroySocket(m_pConInfo->fdClient);
			SAFE_FREE(m_pConInfo->ai_addr)			// ★
			SAFE_FREE(m_pConInfo)
			fRet = TRUE;
		}
	}
	return(fRet);
}

//----------------------------------------------------------------------
// function
//		接続に使用されるインターフェースインデックスを調べる
// parameter
//		addrinfo	*pAddr		[in]接続に使用されるアドレス情報
// return
//		TRUE/FALSE
//----------------------------------------------------------------------
BOOL ShowInterfaceIndex(addrinfo *pAddr)
{
	BOOL			fRet = FALSE;
	sockaddr_in6	addr6;

	if (pAddr->ai_family == AF_INET6)
	{
		memcpy(&addr6, pAddr->ai_addr, sizeof(sockaddr_in6));
		Locate(1, 1, 2);
		fprintf(stderr, "Index:%d\n", addr6.sin6_scope_id);
		fRet = TRUE;
	}
	return(fRet);
}

//----------------------------------------------------------------------
// function
//		HopLimit(IPv6)/TTL(IPv4)を設定する
// parameter
//		SOCKET		fd			[in]ソケット
//		int			iFamily		[in]アドレスファミリ(AF_INET/AF_INET6)
//		int			iHopLimit	[in](HopLimit - 1)台のルータを超えることができる
// return
//		TRUE/FALSE
//----------------------------------------------------------------------
BOOL SetHopLimit(SOCKET fd, int iFamily, int iLimit)
{
	BOOL		fRet = FALSE;
	int			iRet;
	DWORD		dwHoplimit;
	socklen_t	iOptlen;

	if (iFamily == AF_INET6)
	{
		iOptlen = sizeof(dwHoplimit);
		iRet = getsockopt(fd, IPPROTO_IPV6, IPV6_UNICAST_HOPS, (char *)&dwHoplimit, &iOptlen);
		if (iRet == 0)
		{
			Locate(1, 2, 2);
			fprintf(stderr, "HopLimitOrg %d\n", dwHoplimit);
		}
		dwHoplimit = iLimit;
		iOptlen = sizeof(dwHoplimit);
		iRet = setsockopt(fd, IPPROTO_IPV6, IPV6_UNICAST_HOPS, (char *)&dwHoplimit, iOptlen);
		if (iRet == 0)
		{
			iRet = getsockopt(fd, IPPROTO_IPV6, IPV6_UNICAST_HOPS, (char *)&dwHoplimit, &iOptlen);
			if (iRet == 0)
			{
				Locate(1, 3, 2);
				fprintf(stderr, "HopLimitNew %d\n", dwHoplimit);
				fRet = TRUE;
			}
		}
	}
	else if (iFamily == AF_INET)
	{
		iOptlen = sizeof(dwHoplimit);
		iRet = getsockopt(fd, IPPROTO_IP, IP_TTL, (char *)&dwHoplimit, &iOptlen);
		if (iRet == 0)
		{
			Locate(1, 2, 2);
			fprintf(stderr, "TTL %d\n", dwHoplimit);
		}
		dwHoplimit = iLimit;
		iOptlen = sizeof(dwHoplimit);
		iRet = setsockopt(fd, IPPROTO_IP, IP_TTL, (char *)&dwHoplimit, iOptlen);
		if (iRet == 0)
		{
			iRet = getsockopt(fd, IPPROTO_IP, IP_TTL, (char *)&dwHoplimit, &iOptlen);
			if (iRet == 0)
			{
				Locate(1, 3, 2);
				fprintf(stderr, "TTLNew %d\n", dwHoplimit);
				fRet = TRUE;
			}
		}
	}
	return(fRet);
}

//----------------------------------------------
// function
//		メッセージパケットの送信
// parameter
//		char	*pcData		[in]UTF-8データ(NULLターミネートなし)
//		int		iSize		[in]データサイズ
// retun
//		なし
//----------------------------------------------
BOOL	SendMessagePacket(char *pcData, int iSize)
{
	MsgDataRec	*pMsgData = NULL;
	int		iPacketSize = sizeof(HeaderRec) + iSize;		// パケットサイズ
	BOOL		fRet = FALSE;

	pMsgData = (MsgDataRec *)calloc(iPacketSize, sizeof(BYTE));	// パケット全体のエリアを確保

	memcpy(pMsgData->header.bMagicData, MAGIC_STRING, strlen(MAGIC_STRING));
	pMsgData->header.wCommand = htons(CMD_MSG_DATA);
	pMsgData->header.wDataLen = htons(iSize);

	memcpy(pMsgData->bMsgData, pcData, iSize);
	fRet = m_pCSendRecvThread->SetSendData((char *)pMsgData, iPacketSize);

	SAFE_FREE(pMsgData)
	return(fRet);
}
